"""
Utilities for copying metadata to new NeXus files.
"""

from __future__ import annotations

from typing import Any, Dict, List, Tuple

import h5py
import numpy as np
from numpy.typing import ArrayLike

from .. import walk_nxs
from ..beamlines.SSX_chip import Chip, compute_goniometer
from ..nxs_write import calculate_scan_range, create_attributes


def h5str(h5_value: str | np.string_ | bytes) -> str:
    """
    Convert a value returned an h5py attribute to str.

    h5py can return either a bytes-like (numpy.string_) or str object
    for attribute values depending on whether the value was written as
    fixed or variable length. This function collapses the two to str.

    Args:
        h5_value (str | np.string_ | bytes): Original attribute value.

    Returns:
        str: Attribute value collapsed to str.
    """
    if isinstance(h5_value, (np.string_, bytes)):
        return h5_value.decode("utf-8")
    return h5_value


def get_skip_list(nxentry: h5py.Group, skip_obj: List[str]) -> List[str]:
    """
    Get a list of all the objects that should not be copied in the new NeXus file.

    Args:
        nxentry (h5py.Group): NXentry group of a NeXus file.
        skip_obj (List[str]): List of objects that should not be copied.

    Returns:
        skip_list (List[str]): List of "NXclass" objects to skip during copy.
    """
    obj_list = walk_nxs(nxentry)
    skip_list = []
    for obj in obj_list:
        try:
            if nxentry[obj].attrs["NX_class"] in np.string_(skip_obj):
                skip_list.append(obj)
        except Exception:
            pass
    return skip_list


def get_nexus_tree(
    nxs_in: h5py.File,
    nxs_out: h5py.File,
    skip: bool = True,
    skip_obj: List[str] = None,
) -> h5py.Group | None:
    """
    Copy the tree from the original NeXus file. Everything except NXdata is copied to a new NeXus file.
    If skip is False, then the full tree is copied.

    Args:
        nxs_in (h5py.File): Original NeXus file.
        nxs_out (h5py.File): New NeXus file.
        skip (bool, optional): Copy everything but objects in skip_obj, which always include NXdata.
                            Pass False to copy the whole NXentry tree. Defaults to True.
        skip_obj (List[str], optional): List of NX_class objects not to be copied, eg. 'NXdata' or 'NXdetector'.. Defaults to None.

    Returns:
        h5py.Group | None: The group NXentry or nothing if the full file is copied.
    """
    skip_obj = ["NXdata"] if skip_obj is None else skip_obj

    if skip is True:
        nxentry = nxs_out.create_group("entry")
        create_attributes(nxentry, ("NX_class", "default"), ("NXentry", "data"))
        # Copy all of the nexus tree as it is except for the group passed as skip_obj
        skip_list = get_skip_list(nxs_in["entry"], skip_obj)
        # First copy full nxentry
        for k in nxs_in["entry"].keys():
            nxs_in["entry"].copy(k, nxentry)
        # Then delete objects from skip list
        for s in skip_list:
            del nxentry[s]
        return nxentry
    else:
        # Copy everything
        nxs_in.copy("entry", nxs_out)
        return


def identify_tristan_scan_axis(nxs_in: h5py.File) -> Tuple[str | None, Dict[str, Any]]:
    """
    Identify the scan_axis in the NeXus tree of a Tristan collection.

    Return the first data set in the group '/entry/data' that has the attribute
    'transformation_type' equal to 'rotation'.

    Args:
        nxs_in (h5py.File): Tristan NeXus file

    Returns:
        ax (str | None): Name of the scan_axis.
        ax_attrs (Dict[str, Any]): Attributes of the scan_axis dataset.
    """
    nxdata = nxs_in["entry/data"]
    for ax, h5_object in nxdata.items():
        if h5str(h5_object.attrs.get("transformation_type")) == "rotation":
            return ax, dict(h5_object.attrs)
    return None, {}


def convert_scan_axis(nxsample: h5py.Group, nxdata: h5py.Group, ax: str):
    """
    Modify all instances of scan_axis present in NeXus file NXsample group.

    Args:
        nxsample (h5py.Group): NXsample group of NeXus file to be modified.
        nxdata (h5py.Group): NXdata group of NeXus file to be modified.
        ax (str): Name of scan_axis.
    """
    del nxsample["transformations/" + ax]
    nxsample["transformations/" + ax] = nxdata[ax]
    name = (
        "sample_" + ax + "/" + ax if "sam" not in ax else "sample_" + ax[-1] + "/" + ax
    )
    del nxsample[name]
    nxsample[name] = nxdata[ax]


def find_chipmap_in_tristan_nxs(
    nxobj: h5py.File | h5py.Group, loc: str = "entry/source/notes/chipmap"
) -> bool:
    """
    Look for the saved chipmap for a SSX experiment inside a tristan nexus file.

    Args:
        nxobj (h5py.File | h5py.Group): NeXus object to be searched, could be a file or a group.
        loc (str, optional): Location where the chipmap should be saved. Defaults to "entry/source/notes/chipmap".

    Returns:
        bool: Returns True is a chipmap is found, False otherwise.
    """
    obj_list = walk_nxs(nxobj)
    if loc in obj_list:
        return True
    else:
        return False


def compute_ssx_axes(
    nxs_in: h5py.File, nbins: int, rot_ax: str, rot_val: Tuple | List | ArrayLike
) -> Tuple[Dict, Dict, int | None]:
    """
    Computes the positions on chip corresponding to the binned images from a Tristan gri scan collection.
    If multiple windows have been binned into a single image, instead of the sam_(x,y) translation axes values, the number of windows per images will
    be returned and saved in the NeXus file.

    Args:
        nxs_in (h5py.File): File handle for the original Tristan collection NeXus file.
        nbins (int): Number of images.
        rot_ax (str): Rotation axis.
        rot_val (Tuple | List | ArrayLike): Rotation axis start and stop values, as found in the original NeXus.

    Returns:
        OSC, TRANSL, windows_per_bin (Tuple[Dict, Dict, int | None]): Oscillation range, Translation range, number of windows per binned image.
    """
    # Get chipmap: use default chipmap location: /entry/source/notes/chipmap
    blocks = eval(nxs_in["/entry/source/notes/chipmap"][()])

    # Define chip
    chip_info = eval(nxs_in["/entry/source/notes/chip"][()])
    chip = Chip(
        "fastchip",
        num_steps=[chip_info["X_NUM_STEPS"], chip_info["Y_NUM_STEPS"]],
        step_size=[chip_info["X_STEP_SIZE"], chip_info["Y_STEP_SIZE"]],
        num_blocks=[chip_info["X_NUM_BLOCKS"], chip_info["Y_NUM_BLOCKS"]],
        block_size=[chip_info["X_BLOCK_SIZE"], chip_info["Y_BLOCK_SIZE"]],
        start_pos=[
            chip_info["X_START"],
            chip_info["Y_START"],
            chip_info["Z_START"],
        ],
    )

    # Get axis list
    axes_list = [k for k in nxs_in["/entry/sample/transformations"].keys()]
    x_idx = axes_list.index("sam_x")
    y_idx = axes_list.index("sam_y")

    # Rotation values
    OSC = calculate_scan_range(
        [rot_ax], [rot_val[0]], [rot_val[1]], n_images=nbins, rotation=True
    )

    num_blocks = (
        chip.tot_blocks() if list(blocks.values())[0] == "fullchip" else len(blocks)
    )

    # Calculate scan start/end positions on chip
    N_EXP = 1
    if nbins % (num_blocks * chip.tot_windows_per_block()) == 0:
        # All the windows in the selected blocks (full chip or not) have been scanned at least once.
        N_EXP = nbins // (num_blocks * chip.tot_windows_per_block())
        start_pos, end_pos = compute_goniometer(chip, axes_list, blocks=blocks)
        num = (chip.num_steps[1], chip.num_steps[0])
        # Translation values
        TRANSL = {"sam_y": np.array([]), "sam_x": np.array([])}
        for s, e in zip(start_pos.values(), end_pos.values()):
            starts = [
                s[y_idx],
                s[x_idx],
            ]  # because I can't be sure they will be in the right order!
            ends = [
                e[y_idx] - chip.step_size[1],
                e[x_idx] - chip.step_size[0],
            ]  # because scanspec
            transl = calculate_scan_range(
                ["sam_y", "sam_x"], starts, ends, n_images=num
            )
            TRANSL["sam_y"] = np.append(TRANSL["sam_y"], np.round(transl["sam_y"], 3))
            TRANSL["sam_x"] = np.append(TRANSL["sam_x"], np.round(transl["sam_x"], 3))

        if N_EXP > 1:
            OSC = {k: [val for val in v for _ in range(N_EXP)] for k, v in OSC.items()}
            TRANSL = {
                k: [val for val in v for _ in range(N_EXP)] for k, v in TRANSL.items()
            }
        return OSC, TRANSL, None
    else:
        # Hopefully this will not happen..
        # Do not set N_EXP, there should be no repeat here
        # Find out how many consecutive windows are binned together.
        windows_per_bin = (num_blocks * chip.tot_windows_per_block()) // nbins
        return OSC, {}, windows_per_bin



"""Utilities for writing NeXus files for beamlines at Diamond Light Source."""

import argparse

gonioAx_parser = argparse.ArgumentParser(add_help=False)
gonioAx_parser.add_argument("--axes", type=str, nargs="+", help="Axes names.")
gonioAx_parser.add_argument(
    "--ax-start", type=float, nargs="+", help="Axes start positions."
)
gonioAx_parser.add_argument(
    "--ax-inc", type=float, nargs="+", help="Eventual axes increments."
)
gonioAx_parser.add_argument(
    "--ax-end", type=float, nargs="+", help="Eventual axes ends."
)
gonioAx_parser.add_argument("--scan-axis", type=str, help="Identify scan axis.")

detAx_parser = argparse.ArgumentParser(add_help=False)
detAx_parser.add_argument(
    "--det-axes", type=str, nargs="+", help="Detector axes names."
)
detAx_parser.add_argument(
    "--det-start", type=float, nargs="+", help="Detector axes start positions."
)



"""
Create a NeXus file for serial crystallography datasets collected on I24 Eiger 2X 9M detector.
"""
from __future__ import annotations

import glob
import logging
from collections import namedtuple
from pathlib import Path
from typing import Dict, List, Tuple

import h5py
import numpy as np

from .. import get_iso_timestamp, get_nexus_filename, log
from ..nxs_write.NexusWriter import ScanReader, call_writers
from ..nxs_write.NXclassWriters import write_NXdatetime, write_NXentry, write_NXnote
from ..tools.VDS_tools import image_vds_writer
from .I24_Eiger_params import dset_links, eiger9M_params, goniometer_axes, source
from .SSX_chip import Chip, compute_goniometer, read_chip_map

# Define a logger object and a formatter
logger = logging.getLogger("nexgen.I24")

ssx_collect = namedtuple(
    "ssx_collect",
    [
        "visitpath",
        "filename",
        "exp_type",
        "num_imgs",
        "beam_center",
        "detector_distance",
        "start_time",
        "stop_time",
        "exposure_time",
        "transmission",
        "wavelength",
        "flux",
        "pump_status",
        "pump_exp",
        "pump_delay",
        "chip_info",
        "chipmap",
    ],
)

ssx_collect.__doc__ = """Parameters that define a serial collection on I24."""
ssx_collect.visitpath.__doc__ = "Path to colection directory."
ssx_collect.filename.__doc__ = "Root of the filename."
ssx_collect.exp_type.__doc__ = (
    "Experiment being run. Accepted values for I24: extruder, fixed_target, 3Dgridscan."
)
ssx_collect.num_imgs.__doc__ = "Total number of images collected."
ssx_collect.beam_center.__doc__ = "Beam center position, in pixels."
ssx_collect.detector_distance.__doc__ = "Distance between sample and detector, in mm."
ssx_collect.start_time.__doc__ = "Experiment start time."
ssx_collect.stop_time.__doc__ = "Experiment end time."
ssx_collect.exposure_time.__doc__ = "Exposure time, in s."
ssx_collect.transmission.__doc__ = "Attenuator transmission, in %."
ssx_collect.wavelength.__doc__ = "Wavelength of incident beam."
ssx_collect.flux.__doc__ = "Total flux."
ssx_collect.pump_status.__doc__ = "True for a pump-probe experiment, false otherwise."
ssx_collect.pump_exp.__doc__ = "Pump exposure time, in s."
ssx_collect.pump_delay.__doc__ = "Pump delay time, in s."
ssx_collect.chip_info.__doc__ = "For a grid scan, dictionary containing basic chip information. At least it should contain: x/y_start, x/y number of blocks and block size, x/y number of steps and number of exposures."
ssx_collect.chipmap.__doc__ = "Path to the chipmap file corresponding to the experiment, or 'fullchip' indicating that the whole chip is being scanned."

# Define coordinate frame
coordinate_frame = "mcstas"

# Initialize dictionaries
goniometer = goniometer_axes
detector = eiger9M_params
module = {}
beam = {}
attenuator = {}

#
def extruder(
    master_file: Path,
    filename: List[Path],
    num_imgs: int,
    metafile: Path = None,  # Just in case meta is not generated for some reason
    timestamps: Tuple[str] = None,
    pump_info: Dict = None,
):
    """
    Write the NeXus file for extruder collections, pump-probe and not.

    Args:
        master_file (Path): Path to the NeXus file to be written.
        filename (List[Path]): List of paths to data files.
        num_imgs (int): Total number of images passed as a beamline parameter.
        metafile (Path, optional): Path to the Dectris _meta.h5 file. Deafults to None.
        timestamps (Tuple[str], optional): Start and end time of data collection, if known. Defaults to None.
        pump_info (Dict, optional): Details of a pump probe experiment eg. pump exposure time, pump delay, etc. Defaults to None.
    """
    logger.info("Write NeXus file for extruder.")

    goniometer["starts"] = goniometer["ends"] = goniometer["increments"] = [
        0.0,
        0.0,
        0.0,
        0.0,
    ]

    # Get scan range array and rotation axis
    OSC, TRANSL = ScanReader(goniometer, n_images=int(num_imgs))
    del TRANSL

    logger.info("Goniometer information")
    for j in range(len(goniometer["axes"])):
        logger.info(
            f"Goniometer axis: {goniometer['axes'][j]} => {goniometer['starts'][j]}, {goniometer['types'][j]} on {goniometer['depends'][j]}"
        )

    try:
        with h5py.File(master_file, "x") as nxsfile:
            write_NXentry(nxsfile)

            call_writers(
                nxsfile,
                filename,
                coordinate_frame,
                (detector["mode"], num_imgs),
                goniometer,
                detector,
                module,
                source,
                beam,
                attenuator,
                OSC,
                transl_scan=None,
                metafile=metafile,
                link_list=dset_links,
            )

            # Write pump-probe information if requested
            if pump_info:
                logger.info("Write pump information to file.")
                if pump_info["pump_exposure_time"] is None:
                    logger.warning(
                        "Pump exposure time has not been recorded and won't be written to file."
                    )
                if pump_info["pump_delay"] is None:
                    logger.warning(
                        "Pump delay has not been recorded and won't be written to file."
                    )
                loc = "/entry/source/notes"
                write_NXnote(nxsfile, loc, pump_info)

            # Write VDS
            image_vds_writer(nxsfile, (int(num_imgs), *detector["image_size"]))

            if timestamps:
                write_NXdatetime(nxsfile, timestamps)

            logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )


def fixed_target(
    master_file: Path,
    filename: List[Path],
    num_imgs: int,
    chip_info: Dict,
    chipmap: Path | str,
    metafile: Path = None,
    timestamps: Tuple[str] = None,
    pump_info: Dict = None,
):
    """
    Write the NeXus file for fixed target collections, pump probe and not.

    Args:
        master_file (Path): Path to the NeXus file to be written.
        filename (List[Path]):  List of paths to file.
        num_imgs (int): Total number of images passed as a beamline parameter.
        chip_info (Dict): Basic information about the chip in use and collection dynamics.
        chipmap (Path | str): Path to the chipmap file corresponding to the current collection, or string indicating a fullchip scan.
        metafile (Path, optional):  Path to the _meta.h5 file. Defaults to None.
        timestamps (Tuple[str], optional): Start and end time of data collection, if known. See docs for accepted formats. Defaults to None.
        pump_info (Dict, optional): Details of a pump probe experiment eg. pump exposure time, pump delay, etc. Defaults to None.

    Raises:
        ValueError: When the chip information is missing.
    """
    logger.info("Write NeXus file for fixed target.")

    # Check that the chip dict has been passed, raise error is not
    if chip_info is None:
        logger.error("No chip_dict found.")
        raise ValueError(
            "No information about the FT chip has been passed. \
            Impossible to determine scan parameters. NeXus file won't be written."
        )

    chip = Chip(
        "fastchip",
        num_steps=[chip_info["X_NUM_STEPS"][1], chip_info["Y_NUM_STEPS"][1]],
        step_size=[chip_info["X_STEP_SIZE"][1], chip_info["Y_STEP_SIZE"][1]],
        num_blocks=[chip_info["X_NUM_BLOCKS"][1], chip_info["Y_NUM_BLOCKS"][1]],
        block_size=[chip_info["X_BLOCK_SIZE"][1], chip_info["Y_BLOCK_SIZE"][1]],
        start_pos=[
            chip_info["X_START"][1],
            chip_info["Y_START"][1],
            chip_info["Z_START"][1],
        ],
    )

    goniometer["increments"] = [0.0, 0.0, chip.step_size[1], chip.step_size[0]]

    # Read chip map
    blocks = read_chip_map(
        chipmap,
        chip.num_blocks[0],  # chip_info["X_NUM_BLOCKS"][1],
        chip.num_blocks[1],  # chip_info["Y_NUM_BLOCKS"][1],
    )

    # Calculate scan start/end positions on chip
    if list(blocks.values())[0] == "fullchip":
        logger.info("Full chip: all the blocks will be scanned.")
        start_pos, end_pos = compute_goniometer(chip, goniometer["axes"], full=True)
    else:
        logger.info(f"Scanning blocks: {list(blocks.keys())}.")
        start_pos, end_pos = compute_goniometer(chip, goniometer["axes"], blocks=blocks)

    # Iterate over blocks to calculate scan points
    OSC = {"omega": np.array([])}
    TRANSL = {"sam_y": np.array([]), "sam_x": np.array([])}
    for s, e in zip(start_pos.values(), end_pos.values()):
        goniometer["starts"] = s
        goniometer["ends"] = [
            end - inc for end, inc in zip(e, goniometer["increments"])
        ]  # Workaround for scanspec issue (we don't want to write the actual end of the chip)
        osc, transl = ScanReader(
            goniometer,
            n_images=(
                chip.num_steps[1],  # chip_info["Y_NUM_STEPS"][1],
                chip.num_steps[0],  # chip_info["X_NUM_STEPS"][1],
            ),
        )
        OSC["omega"] = np.append(OSC["omega"], osc["omega"])
        TRANSL["sam_y"] = np.append(TRANSL["sam_y"], np.round(transl["sam_y"], 3))
        TRANSL["sam_x"] = np.append(TRANSL["sam_x"], np.round(transl["sam_x"], 3))

    # Log data
    logger.info("Goniometer information")
    for j in range(len(goniometer["axes"])):
        logger.info(
            f"Goniometer axis: {goniometer['axes'][j]} => {goniometer['types'][j]} on {goniometer['depends'][j]}"
        )
    logger.info(f"Oscillation axis: {list(OSC.keys())[0]}.")
    logger.info(f"Grid scan axes: {list(TRANSL.keys())}.")

    if int(chip_info["N_EXPOSURES"][1]) == 1:
        # Check that things make sense
        if num_imgs != len(OSC["omega"]):
            logger.warning(
                f"The total number of scan points is {len(OSC['omega'])}, which does not match the total nu mber of images passed as input {num_imgs}."
            )
            logger.warning(
                "Reset SSX.num_imgs to number of scan points for vds creation"
            )
            tot_imgs = len(OSC["omega"])
        else:
            tot_imgs = num_imgs

        # Write NeXus and VDS
        try:
            with h5py.File(master_file, "x") as nxsfile:
                write_NXentry(nxsfile)

                call_writers(
                    nxsfile,
                    filename,
                    coordinate_frame,
                    (detector["mode"], tot_imgs),
                    goniometer,
                    detector,
                    module,
                    source,
                    beam,
                    attenuator,
                    OSC,
                    transl_scan=TRANSL,
                    metafile=metafile,
                    link_list=dset_links,
                )

                # Write pump-probe information if requested
                if pump_info:
                    logger.info("Write pump information to file.")
                    if pump_info["pump_exposure_time"] is None:
                        logger.warning(
                            "Pump exposure time has not been recorded and won't be written to file."
                        )
                    if pump_info["pump_delay"] is None:
                        logger.warning(
                            "Pump delay has not been recorded and won't be written to file."
                        )
                    # Add pump-repeat to info dictionary
                    pump_info["pump_repeat"] = chip_info["PUMP_REPEAT"][1]
                    loc = "/entry/source/notes"
                    write_NXnote(nxsfile, loc, pump_info)

                # Write VDS
                image_vds_writer(nxsfile, (int(num_imgs), *detector["image_size"]))

                if timestamps:
                    write_NXdatetime(nxsfile, timestamps)

                logger.info(f"The file {master_file} was written correctly.")
        except Exception as err:
            logger.exception(err)
            logger.info(
                f"An error occurred and {master_file} couldn't be written correctly."
            )

    else:
        N = int(chip_info["N_EXPOSURES"][1])
        logger.info(f"Each position has been collected {N} times.")
        pump_repeat = int(chip_info["PUMP_REPEAT"][1])
        logger.info(f"Pump repeat setting: {pump_repeat}")

        # Repeat each position N times
        OSC = {k: [val for val in v for _ in range(N)] for k, v in OSC.items()}
        TRANSL = {k: [val for val in v for _ in range(N)] for k, v in TRANSL.items()}

        # Check that things make sense
        if num_imgs != len(OSC["omega"]):
            logger.warning(
                f"The total number of scan points is {len(OSC['omega'])}, which does not match the total nu mber of images passed as input {num_imgs}."
            )
            logger.warning(
                "Reset SSX.num_imgs to number of scan points for vds creation"
            )
            tot_imgs = len(OSC["omega"])
        else:
            tot_imgs = num_imgs

        # Write NeXus file (for now no VDS)
        try:
            with h5py.File(master_file, "w") as nxsfile:
                write_NXentry(nxsfile)

                call_writers(
                    nxsfile,
                    filename,
                    coordinate_frame,
                    (detector["mode"], tot_imgs),
                    goniometer,
                    detector,
                    module,
                    source,
                    beam,
                    attenuator,
                    OSC,
                    transl_scan=TRANSL,
                    metafile=metafile,
                    link_list=dset_links,
                )

                # Write pump-probe information if requested
                if pump_info:
                    logger.info("Write pump information to file.")
                    if pump_info["pump_exposure_time"] is None:
                        logger.warning(
                            "Pump exposure time has not been recorded and won't be written to file."
                        )
                    if pump_info["pump_delay"] is None:
                        logger.warning(
                            "Pump delay has not been recorded and won't be written to file."
                        )
                    # Add pump-repeat to info dictionary
                    pump_info["pump_repeat"] = chip_info["PUMP_REPEAT"][1]
                    loc = "/entry/source/notes"
                    write_NXnote(nxsfile, loc, pump_info)

                # Write VDS
                # TODO discuss how VDS should be saved. All in one probably not ideal for this application.
                image_vds_writer(nxsfile, (int(num_imgs), *detector["image_size"]))
                # for offset in range(0, tot_imgs, n_windows_in_block*N)
                # image_vds_writer(nxsfile, (int(num_imgs), *detector["image_size"]), offset)

                if timestamps:
                    write_NXdatetime(nxsfile, timestamps)
            logger.info(f"The file {master_file} was written correctly.")
        except Exception as err:
            logger.exception(err)
            logger.info(
                f"An error occurred and {master_file} couldn't be written correctly."
            )


def grid_scan_3D(
    master_file: Path,
    filename: List[Path],
    num_imgs: int,
    chip_info: Dict,
    chipmap: Path | str,
    metafile: Path = None,
    timestamps: Tuple[str] = None,
    pump_info: Dict = None,
):
    """
    Write the NeXus file for 3D grid scans, pump probe and not.

    Args:
        master_file (Path): Path to the NeXus file to be written.
        filename (List[Path]):  List of paths to file.
        num_imgs (int): Total number of images passed as a beamline parameter.
        chip_info (Dict): Basic information about the chip in use and collection dynamics.
        chipmap (Path | str): Path to the chipmap file corresponding to the current collection.
        metafile (Path, optional): Path to the _meta.h5 file. Defaults to None.
        timestamps (Tuple[str], optional): Start and end time of data collection, if known. See docs for accepted formats. Defaults to None.
        pump_info (Dict, optional): Details of a pump probe experiment eg. pump exposure time, pump delay, etc. Defaults to None.

    Raises:
        ValueError: When the chip information is missing.
    """
    logger.info("Write NeXus file for 3D gid scan.")
    # The question here is ... what about rotation?

    logger.info("Write NeXus file for fixed target.")

    # Check that the chip dict has been passed, raise error is not
    if chip_info is None:
        logger.error("No chip_dict found.")
        raise ValueError(
            "No information about the FT chip has been passed. \
            Impossible to determine scan parameters. NeXus file won't be written."
        )

    chip = Chip(
        "fastchip",
        num_steps=[chip_info["X_NUM_STEPS"][1], chip_info["Y_NUM_STEPS"][1]],
        step_size=[chip_info["X_STEP_SIZE"][1], chip_info["Y_STEP_SIZE"][1]],
        num_blocks=[chip_info["X_NUM_BLOCKS"][1], chip_info["Y_NUM_BLOCKS"][1]],
        block_size=[chip_info["X_BLOCK_SIZE"][1], chip_info["Y_BLOCK_SIZE"][1]],
        start_pos=[
            chip_info["X_START"][1],
            chip_info["Y_START"][1],
            chip_info["Z_START"][1],
        ],
    )
    # Read chip map
    blocks = read_chip_map(
        chipmap,
        chip.num_blocks[0],  # chip_info["X_NUM_BLOCKS"][1],
        chip.num_blocks[1],  # chip_info["Y_NUM_BLOCKS"][1],
    )
    print(blocks)


def write_nxs(**ssx_params):
    """
    Gather all parameters from the beamline and call appropriate writer function for serial crystallography.
    """
    # Get info from the beamline
    SSX = ssx_collect(
        visitpath=Path(ssx_params["visitpath"]).expanduser().resolve(),
        filename=ssx_params["filename"],  # Template: test_##
        exp_type=ssx_params["exp_type"],
        num_imgs=float(ssx_params["num_imgs"]),
        beam_center=ssx_params["beam_center"],
        detector_distance=ssx_params["det_dist"],
        start_time=ssx_params["start_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if ssx_params["start_time"]
        else None,  # This should be datetiem type
        stop_time=ssx_params["stop_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if ssx_params["stop_time"]
        else None,  # idem.
        exposure_time=ssx_params["exp_time"],
        transmission=ssx_params["transmission"],
        wavelength=ssx_params["wavelength"],
        flux=ssx_params["flux"],
        pump_status=ssx_params["pump_status"],
        pump_exp=ssx_params["pump_exp"],
        pump_delay=ssx_params["pump_delay"],
        chip_info=None
        if ssx_params["exp_type"] == "extruder"
        else ssx_params[
            "chip_info"
        ],  # ssx_params["chip_info"] if ssx_params["chip_info"] else None,
        chipmap=None
        if ssx_params["exp_type"] == "extruder"
        else ssx_params["chipmap"],  # It could be a path or just a flag to fullchip
        # else Path(ssx_params["chipmap"]).expanduser().resolve(),
    )

    logfile = SSX.visitpath / "I24_nxs_writer.log"
    # Configure logging
    log.config(logfile.as_posix())

    # Add to dictionaries
    detector["starts"] = [SSX.detector_distance]
    detector["exposure_time"] = SSX.exposure_time
    detector["beam_center"] = SSX.beam_center

    attenuator["transmission"] = SSX.transmission

    beam["wavelength"] = SSX.wavelength
    beam["flux"] = SSX.flux

    logger.info(f"Current collection directory: {SSX.visitpath}")
    # Find metafile in directory and get info from it
    try:
        metafile = [
            f for f in SSX.visitpath.iterdir() if SSX.filename + "_meta" in f.as_posix()
        ][0]
        logger.info(f"Found {metafile} in directory.")
    except IndexError:
        logger.warning(
            "No _meta.h5 file found in directory. External links in the NeXus file will be broken."
        )
        logger.error(
            "Missing metadata, unable to write NeXus file. Please use command line tool."
        )
        # TODO add instructions for using command line tool

    module["fast_axis"] = detector.pop("fast_axis")
    module["slow_axis"] = detector.pop("slow_axis")
    # Set value for module_offset calculation.
    module["module_offset"] = "1"

    # Find datafiles
    filename_template = (
        metafile.parent / metafile.name.replace("meta", f"{6*'[0-9]'}")
    ).as_posix()
    # if meta else (SSX.visitpath / SSX.filename).as_posix() + f"_{6*'[0-9]'}.h5"
    filename = [
        Path(f).expanduser().resolve() for f in sorted(glob.glob(filename_template))
    ]

    # Add some information to logger
    logger.info("Creating a NeXus file for %s ..." % metafile.name)
    # Get NeXus filename
    master_file = get_nexus_filename(filename[0])
    logger.info("NeXus file will be saved as %s" % master_file)

    logger.info("Detector information")
    logger.info(f"{detector['description']}")
    logger.info(
        f"Sensor made of {detector['sensor_material']} x {detector['sensor_thickness']}"
    )
    logger.info(
        f"Detector is a {detector['image_size'][::-1]} array of {detector['pixel_size']} pixels"
    )
    for k in range(len(detector["axes"])):
        logger.info(
            f"Detector axis: {detector['axes'][k]} => {detector['starts'][k]}{detector['units'][k]}, {detector['types'][k]} on {detector['depends'][k]}"
        )

    logger.info(f"Recorded beam center is: {SSX.beam_center}.")

    # Get timestamps in the correct format
    timestamps = (
        get_iso_timestamp(SSX.start_time),
        get_iso_timestamp(SSX.stop_time),
    )
    logger.info(f"Timestamps recorded: {timestamps}")

    if SSX.pump_status == "true":
        logger.info("Pump status is True.")
        pump_info = {}
        pump_info["pump_exposure_time"] = SSX.pump_exp if SSX.pump_exp else None
        logger.info(f"Recorded pump exposure time: {SSX.pump_exp}")
        pump_info["pump_delay"] = SSX.pump_delay if SSX.pump_delay else None
        logger.info(f"Recorded pump delay time: {SSX.pump_delay}")
    else:
        pump_info = None

    # Call correct function for the current experiment
    if SSX.exp_type == "extruder":
        extruder(
            master_file, filename, int(SSX.num_imgs), metafile, timestamps, pump_info
        )
    elif SSX.exp_type == "fixed_target":
        # Define chipmap if needed
        chipmapfile = (
            None
            if SSX.chipmap == "fullchip"
            else Path(SSX.chipmap).expanduser().resolve()
        )

        fixed_target(
            master_file,
            filename,
            int(SSX.num_imgs),
            SSX.chip_info,
            chipmapfile,
            metafile,
            timestamps,
            pump_info,
        )
    elif SSX.exp_type == "3Dgridscan":
        # Define chipmap if needed
        chipmapfile = (
            None
            if SSX.chipmap == "fullchip"
            else Path(SSX.chipmap).expanduser().resolve()
        )

        grid_scan_3D(
            master_file,
            filename,
            int(SSX.num_imgs),
            SSX.chip_info,
            chipmapfile,
            metafile,
            timestamps,
            pump_info,
        )

    logger.info("*EOF*\n")


# # Example usage
# if __name__ == "__main__":
#     import sys
#     from datetime import datetime

#     write_nxs(
#         visitpath=sys.argv[1],
#         filename=sys.argv[2],
#         exp_type="extruder",
#         num_imgs=2450,
#         beam_center=[1590.7, 1643.7],
#         det_dist=0.5,
#         # start_time=None,
#         # stop_time=None,
#         start_time=datetime.now(),
#         stop_time=datetime.now(),
#         exp_time=0.002,
#         transmission=1.0,
#         wavelength=0.649,
#         flux=None,
#         pump_status="true",  # this is a string on the beamline
#         pump_exp=None,
#         pump_delay=None,
#         chip_info=None,
#         chipmap=None,
#     )



"""
Create a NeXus file for serial crystallography datasets collected on Tristan10M detector either on I19-2 or I24 beamlines.
"""

import logging
from collections import namedtuple
from pathlib import Path

import h5py

from .. import get_iso_timestamp, get_nexus_filename, log
from ..nxs_write.NexusWriter import call_writers
from ..nxs_write.NXclassWriters import write_NXdatetime, write_NXentry, write_NXnote
from .I19_2_params import source, tristan10M_params

# Define a logger object and a formatter
logger = logging.getLogger("nexgen.I19-2_ssx")

ssx_tr_collect = namedtuple(
    "ssx_collect",
    [
        "visitpath",
        "filename",
        "location",
        "beam_center",
        "detector_distance",
        "start_time",
        "stop_time",
        "exposure_time",
        "transmission",
        "wavelength",
        "chipmap",
        "chip_info",
    ],
)

ssx_tr_collect.__doc__ = (
    """Parameters that define a serial collection using a Tristan detector."""
)
ssx_tr_collect.visitpath.__doc__ = "Path to colection directory."
ssx_tr_collect.filename.__doc__ = "Root of the filename."
ssx_tr_collect.location.__doc__ = "Beamline on which the experiment is being run."
ssx_tr_collect.beam_center.__doc__ = "Beam center position, in pixels."
ssx_tr_collect.detector_distance.__doc__ = (
    "Distance between sample and detector, in mm."
)
ssx_tr_collect.start_time.__doc__ = "Experiment start time."
ssx_tr_collect.stop_time.__doc__ = "Experiment end time."
ssx_tr_collect.exposure_time.__doc__ = "Exposure time, in s."
ssx_tr_collect.transmission.__doc__ = "Attenuator transmission, in %."
ssx_tr_collect.wavelength.__doc__ = "Wavelength of incident beam."
ssx_tr_collect.chipmap.__doc__ = "Chipmap or block list for grid scan."
ssx_tr_collect.chip_info.__doc__ = "For a grid scan, dictionary containing basic chip information. At least it should contain: x/y_start, x/y number of blocks and block size, x/y number of steps and number of exposures."

# Define coordinate frame
coordinate_frame = "mcstas"

# Initialize dictionaries
goniometer = {}
detector = tristan10M_params
module = {}
beam = {}
attenuator = {}


def write_nxs(**ssx_params):
    """
    Gather all parameters from the beamline and call the NeXus writers.
    """
    # Get info from the beamline
    SSX_TR = ssx_tr_collect(
        visitpath=Path(ssx_params["visitpath"]).expanduser().resolve(),
        filename=ssx_params["filename"],
        location=ssx_params["location"],
        beam_center=ssx_params["beam_center"],
        detector_distance=ssx_params["det_dist"],
        start_time=ssx_params["start_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if ssx_params["start_time"]
        else None,  # This should be datetiem type
        stop_time=ssx_params["stop_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if ssx_params["stop_time"]
        else None,  # idem.
        exposure_time=ssx_params["exp_time"],
        transmission=ssx_params["transmission"],
        wavelength=ssx_params["wavelength"],
        chipmap=ssx_params["chipmap"] if ssx_params["chipmap"] else None,
        chip_info=ssx_params["chip_info"] if ssx_params["chip_info"] else None,
    )

    logfile = SSX_TR.visitpath / "TristanSSX_nxs_writer.log"
    # Configure logging
    log.config(logfile.as_posix())

    logger.info(
        f"Start NeXus File Writer for time-resolved SSX on beamline {source['beamline_name']} at DLS."
    )

    # Add to dictionaries
    # Detector
    # If location is I24, two_theta is not present
    detector["starts"] = (
        [0.0, SSX_TR.detector_distance]
        if "I19" in SSX_TR.location
        else [SSX_TR.detector_distance]
    )
    if "I24" in SSX_TR.location:
        detector["axes"] = ["det_z"]
        detector["types"] = ["translation"]
        detector["units"] = ["mm"]
        detector["depends"] = ["."]
        detector["vectors"] = [0, 0, 1]
        detector["increments"] = [0.0]
    detector["exposure_time"] = SSX_TR.exposure_time
    detector["beam_center"] = SSX_TR.beam_center

    # Module
    module["fast_axis"] = detector.pop("fast_axis")
    module["slow_axis"] = detector.pop("slow_axis")
    # goniometer, detector, module = read_params_from_json()
    # Set value for module_offset calculation.
    module["module_offset"] = "1"

    # Attenuator
    attenuator["transmission"] = SSX_TR.transmission

    # Beam
    beam["wavelength"] = SSX_TR.wavelength
    beam["flux"] = None

    # Goniometer
    if "I19" in SSX_TR.location:
        from .I19_2_params import goniometer_axes
    elif "I24" in SSX_TR.location:
        from .I24_Eiger_params import goniometer_axes

    for k, v in goniometer_axes.items():
        goniometer[k] = v

    l = len(goniometer["axes"])
    goniometer["starts"] = goniometer["ends"] = goniometer["increments"] = l * [0.0]

    # Get rotation scan range array and axis
    osc_axis = "phi" if "I19" in SSX_TR.location else "omega"
    osc_range = (0.0, 0.0)

    OSC = {osc_axis: osc_range}

    logger.info("Goniometer information")
    for j in range(len(goniometer["axes"])):
        logger.info(
            f"Goniometer axis: {goniometer['axes'][j]} => {goniometer['starts'][j]}, {goniometer['types'][j]} on {goniometer['depends'][j]}"
        )
    logger.info("Detector information")
    logger.info(f"{detector['description']}")
    logger.info(
        f"Sensor made of {detector['sensor_material']} x {detector['sensor_thickness']}"
    )
    logger.info(
        f"Detector is a {detector['image_size'][::-1]} array of {detector['pixel_size']} pixels"
    )
    for k in range(len(detector["axes"])):
        logger.info(
            f"Detector axis: {detector['axes'][k]} => {detector['starts'][k]}, {detector['types'][k]} on {detector['depends'][k]}"
        )

    # Get timestamps in the correct format
    timestamps = (
        get_iso_timestamp(SSX_TR.start_time),
        get_iso_timestamp(SSX_TR.stop_time),
    )
    logger.info(f"Timestamps recorded: {timestamps}")

    logger.info(f"Current collection directory: {SSX_TR.visitpath}")
    # Find metafile in directory and get info from it
    metafile = [
        f
        for f in SSX_TR.visitpath.iterdir()
        if SSX_TR.filename + "_meta" in f.as_posix()
    ][0]
    logger.info(f"Found {metafile} in directory.")

    # Add some information to logger
    logger.info("Creating a NeXus file for %s ..." % metafile.name)
    # Get NeXus filename
    master_file = get_nexus_filename(metafile)
    logger.info("NeXus file will be saved as %s" % master_file)

    try:
        with h5py.File(master_file, "x") as nxsfile:
            write_NXentry(nxsfile)

            if timestamps[0]:
                write_NXdatetime(nxsfile, (timestamps[0], None))

            call_writers(
                nxsfile,
                [metafile],
                coordinate_frame,
                (
                    detector["mode"],
                    None,
                ),  # since it's event mode, don't need event number or chunk number here
                goniometer,
                detector,
                module,
                source,
                beam,
                attenuator,
                OSC,
                transl_scan=None,
                metafile=metafile,  # Since there are no links, this could also be None
                link_list=None,
            )

            # Save chipmap (list of city blocks)
            if SSX_TR.chipmap:
                # Assuming 8x8 fast chip
                from .SSX_chip import read_chip_map

                chip = read_chip_map(SSX_TR.chipmap, 8, 8)
                mapping = {"chipmap": str(chip)}
                logger.info(
                    f"Chipmap read from {SSX_TR.chipmap}, saving in '/entry/source/notes/chipmap'."
                )
                write_NXnote(nxsfile, "/entry/source/notes/", mapping)
                # To read this: eval(dset[()])
            # Save chip info if passed. If not save I24-like chip info plus warning message
            logger.info("Save chip information in /entry/source/notes/chip")
            if SSX_TR.chip_info:
                # Make chip info more readable
                chip_info = {k: v[1] for k, v in SSX_TR.chip_info.items()}
                chipdef = {"chip": str(chip_info)}
                write_NXnote(nxsfile, "/entry/source/notes/", chipdef)
            else:
                logger.warning(
                    f"Dictionary containing chip info was not passed to the writer."
                    "The following values will be written as default: "
                    "x/y_num_blocks = 8 \n x/y_block_size = 3.175 \n x/y_num_steps = 20 \n x/y_step_size = 0.125"
                )
                chip_info = {
                    "X_NUM_STEPS": 20,
                    "Y_NUM_STEPS": 20,
                    "X_STEP_SIZE": 0.125,
                    "Y_STEP_SIZE": 0.125,
                    "X_START": 0,
                    "Y_START": 0,
                    "Z_START": 0,
                    "X_NUM_BLOCKS": 8,
                    "Y_NUM_BLOCKS": 8,
                    "X_BLOCK_SIZE": 3.175,
                    "Y_BLOCK_SIZE": 3.175,
                }
                chipdef = {"chip": str(chip_info)}
                write_NXnote(nxsfile, "/entry/source/notes", chipdef)

            if timestamps[1]:
                write_NXdatetime(nxsfile, (None, timestamps[1]))
            logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )


# # Example usage
# if __name__ == "__main__":
#     from datetime import datetime

#     write_nxs(
#         visitpath=sys.argv[1],
#         filename=sys.argv[2],
#         location="I19",
#         beam_center=[1590.7, 1643.7],
#         det_dist=0.5,
#         start_time=datetime.now(),
#         stop_time=None,
#         exp_time=0.002,
#         transmission=1.0,
#         wavelength=0.649,
#         chipmap=None,
#         chip_info=None,
#     )
